
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

module Tomte::Protocol::Endpoint
  class Error < StandardError; end


  #
  #
  #
  #
  class Adapter

    class ConnectionError < StandardError
    end
    class AuthenticationError < StandardError
    end

    DEFAULT_ENVELOPER = nil

    attr_reader :identity, :name
    attr_reader :log

    attr_accessor :enveloper

    attr_reader :client
    attr_reader :options

    #
    # Gives a new endpoint,
    # takes several options in a hash, including a list of AMQP options
    #
    #  Tomte::Protocol::Endpoint.new(:name => "my new endpoint",
    #                                :uri => "amqp://user:password@host.name/vhost
    #                                :serialization => :json,
    #
    # Or, the specification of a bus:
    #
    #   Tomte::Protocol::Endpoint.new(:bus => :system)
    #
    # which must be configured from the Tomte::Config
    #
    # It may include an entire enveloper
    #
    #   Tomte::Protocol::Endpoint.new(:bus => :system, :enveloper => Tomte::Protocol::Enveloper::Fancy.new())
    #
    # @param [String] name a string identifying this adapter
    # @param [Symbol] serialization_format :marshal, :json, :yaml ...
    # @param [Symbol] uri amqp URI for the broker
    # @param [Boolean] logging 'true'
    # @param [Integer] timeout 10
    # @param [Symbol] bus :system
    # @param [Tomte::Protocol::Enveloper] enveloper an initialized enveloper
    #
    #

    def initialize(options = {})

      # ensure the framework is configured
      Tomte::Protocol.configure!

      @log = options.fetch(:logger, Tomte.logger)

      # allows to use a symbol
      options = {:bus => options} if options.is_a? Symbol

      # allows to use a string
      begin
        options = parse_uri(options)
      rescue ArgumentError
        options = {:bus => options}
      end if options.is_a? String


      if(busname = options.delete(:bus))
        begin
          config = Tomte::Config.protocol[:buses].fetch(busname).dup
          options.merge! config if config
          tomte_debug do
            log.debug "- Endpoint, bus configuration:"
            log.debug config.ai.subindent
          end
          options[:name] ||= "endpoint for bus '#{busname}'"
        rescue KeyError
          raise Error, "The requested bus '#{busname}' is not configured."
        end
      end

      if(uri = options.delete(:uri))
        begin
          options.merge! parse_uri(uri.chomp('/')).reject{|k,v| v.nil?}
          options[:name] ||= "endpoint to '#{uri}'"
        rescue ArgumentError => e
          raise Error, e.message
        end
      end

      @name = options[:name] || "default endpoint"
      @identity =  options[:identity] || Tomte.config.local[:identity] ||
        Tomte.random_id

      log.debug "- Initialized Endpoint '#{self}'."
      options[:serialization] ||= :bson

      set_global_enveloper options[:enveloper]

      @options = options
    end

    def connect!(&block)

      block.call(self) if block_given?
    end


    def to_s
      "#{self.class}:[#{self.name} (Id: #{self.identity})]"
    end

    def shutdown!

    end

    def settings
      client.settings if client
    end



    #
    # true if this adapter is connected
    #
    def connected?
      !!(@client && @client.connected?)
    end

    private
    #######

    def set_global_enveloper(selected_enveloper)
      case selected_enveloper
      when nil
        DEFAULT_ENVELOPER ? DEFAULT_ENVELOPER.new : nil
      when Class
        @enveloper = selected_enveloper.new
      else
        @enveloper = selected_enveloper
      end

    end

    def parse_uri(uri)
      raise NotImplementedError
    end
  end
end