
# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'amqp'
#require 'amqp/client'
#require 'amqp/extensions/rabbitmq'

require 'thread'

module Tomte::Protocol::Endpoint


  #
  # = AMQP bus adapter
  #
  class Async < Tomte::Protocol::Endpoint::Adapter
    include Entities

    def initialize(options = {})


      super
      @options[:on_tcp_connection_failure] ||= method(:generic_on_tcp_connection_failure)
      @options[:on_possible_authentication_failure] ||= method(:generic_on_possible_authentication_failure)
    end


    #
    # XXX not sure
    #
    def start!(&block)
      @client = AMQP.start(@options) do |session|
        session.logger = log
        block.call self if block_given?
      end
      self
    end


    def connect!(&block)
      ensure_reactor_running!

      # FIXME
      # documentation says I should do like this, but the logger property
      # doesn't get mixed in before the client is booted up during connect
      #
      @options[:auto_recovery] = true
      @options[:heartbeat] ||= 30

      EM.next_tick do

        tomte_debug do
          log.debug "- Endpoint options:"
          log.debug @options.reject{ |k| [:enveloper ].include?(k) }.ai.subindent
        end

        @client = AMQP.connect(@options) do |session|

          # see above note
          client.logger = log

          session.client_properties.merge!(
            :product => "Tomte",
            :information => 'http://hobbes.windmill.it/tomte',
            :version => Tomte::VERSION
          )

          log.debug "+ Endpoint '#{name}' is connected to broker, Automatic recovery: #{session.auto_recovering?}, hb: #{@options[:heartbeat]}s"
          block.call(self) if block
        end

        @client.on_error &method(:error)
        @client.on_tcp_connection_loss &method(:tcp_connection_loss)
        @client.on_skipped_heartbeats &method(:skipped_heartbeats)
        @client.on_open { log.debug "+ Endpoint '#{name}' session open."}

      end
      self
    end

    def disconnect!(&block)
      begin
        @client.disconnect do
          log.debug "- Endpoint '#{name}', disconnected."
          block.call if block_given?
        end
      rescue
        log.error "- Endpoint '#{name}', error in disconnect!: #{$!.to_s}".red
        log.debug "\n |  #{$!.class.name} (#{$!.to_s})\n |  #{$!.backtrace.join("\n |  ")}".blue
        block.call if block_given?
      end
    end

    def sync?
      false
    end

    #
    # so that we can get the async endpoint regardless of having a sync or async one
    #
    def async
      self
    end

    def new_channel(options = {}, &block)
      Tomte::Protocol::Channel.new(client, nil,
                                   options.merge(:enveloper => enveloper), &block)

    end

    #
    # sets the trace flags on
    #
    def trace!
      Tomte.config.protocol[:trace] = true
    end


    #
    #
    #
    def ensure_connected(&block)
      if @client
        block.call
      else
        connect! &block
      end
    end

    private
    #######

    def skipped_heartbeats(conn, settings)
      log.error "= Endpoint '#{name}', detected skipped heartbeats...".important!
    end

    def tcp_connection_loss(conn, settings)
      log.error "= Endpoint '#{name}', tcp connection loss. reconnecting...".important!
      conn.reconnect(false, 10)
    end

    def error(conn, connection_close)
      log.error "= Endpoint '#{name}', [connection.close] Reply code = #{connection_close.reply_code}, reply text = #{connection_close.reply_text}"

      if connection_close.reply_code == 320
        log.error  "+ Endpoint '#{name}', [connection.close] Setting up a periodic reconnection timer..."

        # every 10 seconds
        conn.periodically_reconnect(10)
      end
    end

    def generic_on_tcp_connection_failure(settings)
      log.error "= Endpoint '#{name}' failed to connect to the broker.".important!
      log.debug settings.ai.subindent
      raise ConnectionError.new "= Endpoint '#{name}' failed to connect to the broker."
    end


    def generic_on_possible_authentication_failure(settings)
      log.error "= Endpoint '#{name}' failed to authenticate to the broker.".important!
      log.debug settings.ai.subindent
      raise AuthenticationError.new "= Endpoint '#{name}' failed to authenticate to the broker."
    end


    def stray_exception(e)
       raise e if e.is_a? SystemExit

       if log
         log.error "= ====== unhandled exception [#{e.class.name}] ======".important!
         log.exception(e)
       else
         raise e
       end

       shutdown!
     end


    def shutdown!
      super
    end


    def parse_uri(uri)
      AMQ::URI.parse(uri)
    end

    def ensure_reactor_running!

      unless EM.reactor_running?

        log.debug "+ Async endpoint is starting the reactor."
        Thread.abort_on_exception = true

        thr = Thread.current
        Thread.new { EM.run { EM.next_tick { thr.wakeup } } }
        Thread.stop
        log.debug "+ Async endpoint has done starting the reactor."

        EM.error_handler { |e| stray_exception e }
      end


#      Thread.new { EM.run } unless EM.reactor_running?
#      while not EM.reactor_running?; end
    end

  end
end
