module Tomte::Protocol

  #
  # base class for all protocol errors
  #
  class Error < StandardError
    extend Forwardable

    def initialize(message = nil, options = {})
      super(message)
      @message = message if message
    end
  end

  #
  # An request has timed out before compelting
  #
  class Timeout < Error
    def initialize(message = "Request Timed Out")

      super
    end
  end


  #
  # The request failed because of a protocol error
  # (ie: it most probably caused an AMQP exception )
  #
  class ProtocolFailure < Error

    attr_accessor :transport_exception
    def_delegators :@transport_exception, :reply_code, :reply_text

    def initialize(message = "A Protocol Failure Occurred", options = {})
      super

      @transport_metadata  = options[:transport_metadata]
      @transport_exception = options[:transport_exception]
    end

    def type
      @transport_exception.class.name
    end


    def to_s(detailed = false)
      res = @message + " [#{type}: (#{reply_code}: #{reply_text})]\n"

      if detailed
        details = {}
        if @transport_exception
          details[:exception] =
            [:reply_code, :reply_text].inject_hash do |res, i|
              res[i] = @transport_exception.__send__(i)
            end
        end

        details[:metadata] = @transport_metadata
        res += details.ai.subindent
      end

      res
    end
  end

  #
  # An error has occurred in the request during the transport of the messages
  # (ie: the message was unroutable or it had no acceptors)
  #
  class TransportFailure < Error

    attr_accessor :transport_metadata
    attr_accessor :transport_exception

    def_delegators :@transport_exception, :reply_code, :reply_text

    def initialize(message = "A Transport Failure Occurred", options = {})
      super

      @transport_metadata  = options[:transport_metadata]
      @transport_exception = options[:transport_exception]
    end

    def type
      @transport_exception.class.name
    end

    def to_s(detailed = false)
      res = @message + " [#{type}: (#{reply_code}: #{reply_text})]\n"

      if detailed
        details = {}
        if @transport_exception
          details[:exception] =
            [:exchange, :routing_key, :reply_code, :reply_text].inject_hash do |res, i|
              res[i] = @transport_exception.__send__(i)
            end
        end

        details[:metadata] = @transport_metadata
        res += details.ai.subindent
      end
    res
    end
  end

  #
  # Either the request, or the response was not in the correct format
  #
  class FormatException < Error
    attr_accessor :format_exception
    attr_accessor :metadata
    attr_accessor :payload

    def initialize(message = "A Transport Failure Occurred", options = {})
      super

      @metadata, @payload, @format_exception = [
          options[:metadata],
          options[:payload],
          options[:format_exception]]

    end

    def to_s
      "#{@message} \n  [#{@format_exception.to_s}]"
    end
  end

  #
  # an error has occurred on the other side of the dialog
  #
  class RemoteException < Error

    attr_accessor :remote
    def_delegators :@remote, :remote_message, :caption

    def initialize(message = "A remote exception occurred", options = {})
      super

      @remote = OpenStruct.new(options[:remote]).freeze
    end

    def type
      @remote._type
    end

    def to_s
      "#{@message} \n  [#{type}: #{remote_message}]"
    end
  end

  #
  # an error has occurred on the other side of the RPC request
  # and it contains the remote exception
  #
  class RPCRemoteException < Error
    attr_accessor :remote

    def initialize(message = "A remote exception occurred during the RPC request", options = {})
      super

      @remote = options[:remote]

      if remote.is_a?(Exception) && remote.backtrace.respond_to?(:each)
        remote.set_backtrace(remote.backtrace.map {|v| "[remote]#{v}" })
      end
    end

    def type
      remote.class.name
    end

    def remote_message
      remote.message
    end
    alias_method :caption, :remote_message

    def to_s
      "#{@message} \n  [#{remote.to_s}]"
    end

  end
end