# Tomte, a framework for distributed agents.
# Copyright (c) 2010-2012, Lele Forzani, windmillmedia. All Rights Reserved.
#
# Author:: Lele Forzani <lele@windmill.it>
# License:: You can redistribute it and/or modify it under the terms of the LICENSE file.

require 'thread'

module Tomte::Protocol

  class WaitCondition

    def initialize(options = {}, &callback)

      @forced = options[:forced]

      @callback = callback
      @queue = ::Queue.new if synchronized?
    end

    def synchronized?
      @forced ||
          @synchronized ||= (!@callback && !EM.reactor_thread?)
    end

    def wait
      data = @queue.pop if @queue
      exception(data) if data.is_a?(Exception)
      data
    end

    def return_raise(exception)
      # maybe one day do something cleverer
      self.return exception
    end

    def return(*status)
      @callback.call *status if @callback
      @queue << status.last if @queue
    end

    def exception(e)
      e.backtrace.try(:empty?) ? raise(e, e.try(:message), caller(2)) : raise(e)
    end

  end

  module Synchronization

    class LockCondition < StandardError; end

    module Channel
      extend ActiveSupport::Concern

        included do
          include Tomte::Protocol::Synchronization
        end


      def initialize(connection = nil, id = self.class.next_channel_id, options = {}, &block)
        resync(block, options) do |syn|
          super {|*res| syn.return *res }
        end
      end
    end


    module Exchange
      extend ActiveSupport::Concern

      included do
        include Tomte::Protocol::Synchronization
      end


      def initialize(channel, type, name, opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def publish(payload, options = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def delete(opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

    end


    module Queue
      extend ActiveSupport::Concern

      included do
        include Tomte::Protocol::Synchronization
      end


      def initialize(channel, name = AMQ::Protocol::EMPTY_STRING, opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def bind(exchange, opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def unbind(exchange, opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def delete(opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def purge(opts = {}, &block)
        resync(block, opts) do |syn|
          super {|*res| syn.return *res }
        end
      end

      def pop(opts = {}, &block)
        resync(block, opts) do |syn|
          super {|headers, payload| syn.return *[headers, payload] }
        end
      end

      def fetch(opts = {}, &block)
        resync(block, opts) do |syn|
          subscribe(opts) do |headers, payload|
            unsubscribe
            syn.return *[headers, payload]
          end
        end
      end

    end


    #
    #
    #
    #
    def resync(callback = nil, options = {}, &block)
      raise LockCondition, 'Performing Synchronization on the reactor thread?' if EM.reactor_thread?

      wc = WaitCondition.new(:forced => true, &callback)

      if EM.reactor_running?
        EM.next_tick do
          block.call(wc)
        end
      else
        block.call(wc)
      end

      wc.wait
    end
  end

end





